/** Estimate semi-structural parameters of hybrid NKPC model
	using Jeffreys and g-priors, with A0 calibrated to Del Negro Schorheide Smets and Wouters (2007) (DSSW) priors.
**/

#include <oxstd.h>
#include <oxdraw.h>
#include <oxprob.h>
#import <maximize>
#import <database>

decl g_A, g_OmegaInv, g_XX;

structural2semi(const vp)
/**	Mapping from structural to semistructural parameters
**/
{
	decl calvo, beta, varrho;					// structural parameters
	calvo = vp[0]; beta = vp[1]; varrho = vp[2];
	decl arg= 1+varrho*beta;						// to save calculations
	decl lambda, gammaf, gammab;				// semistructural params
	lambda 	= (1-calvo)*(1-calvo*beta)/arg/calvo;
	gammaf 	= beta/arg;
	gammab 	= varrho/arg;
	return {lambda, gammaf, gammab};
}
getsolution(const vp, const amA)
/**	Analytical solution, a la Kurmann (2007)	**/
{
	decl lambda, gammaf, gammab;				// semistructural params
	lambda = vp[0]; gammaf = vp[1]; gammab = vp[2];
	decl a_pp1 = vp[3], a_px1 = vp[4], a_pp2 = vp[5], a_px2 = vp[6];					// coefficients in the reduced-form equation for pi
	decl denom = lambda + gammaf*a_px1;
	decl a_xp1 = ((1-gammaf*a_pp1)*a_pp1 - gammaf * a_pp2 - gammab)/denom;
	decl a_xx1 = ((1-gammaf*a_pp1)*a_px1 - gammaf * a_px2)/denom;
	decl a_xp2 = (1-gammaf*a_pp1)*a_pp2/denom;
	decl a_xx2 = (1-gammaf*a_pp1)*a_px2/denom;
	amA[0] = (a_pp1 | a_px1 | a_pp2 | a_px2) ~ (a_xp1 | a_xx1 | a_xp2 | a_xx2);
	return !isnan(amA[0]);
}

JacvecAwrtTheta(const avJac, const vp)
/**	Jacobian of vec(A(theta))	**/
{
	decl mJac, iresult;				 	// Jacobian of the transformation from structural to reduced form
	decl lambda, gammaf, gammab;				// semistructural params
	lambda = vp[0]; gammaf = vp[1]; gammab = vp[2];
	decl a_pp1 = vp[3], a_px1 = vp[4], a_pp2 = vp[5], a_px2 = vp[6];					// coefficients in the reduced-form equation for pi
	decl mA;
	getsolution(vp, &mA);			// obtain reduced form parameters
	decl a_xp1 = mA[0][1], a_xx1 = mA[1][1], a_xp2 = mA[2][1], a_xx2 = mA[3][1];
	decl denom = lambda + gammaf*a_px1;

	mJac = zeros(4,3) ~ unit(4) |
			-a_xp1/denom ~ -(lambda*a_pp1^2+a_px1*a_pp1-gammab*a_px1+lambda*a_pp2)/denom^2 ~ -1/denom ~ (1-2*gammaf*a_pp1)/denom ~ -gammaf*a_xp1/denom ~ -gammaf/denom ~ 0
	 	|	-a_xx1/denom ~ -(a_px1^2+lambda*a_pp1*a_px1+lambda*a_px2)/denom^2 ~ 0 ~ -gammaf*a_px1/denom ~ (a_px2*gammaf^2-lambda*a_pp1*gammaf+lambda)/denom^2 ~ 0 ~ -gammaf/denom
		|   -a_xp2/denom ~ -(lambda*a_pp1+a_px1)*a_pp2/denom^2 ~ 0 ~ -gammaf*a_pp2/denom ~ -gammaf*a_xp2/denom ~ (1-gammaf*a_pp1)/denom ~ 0
		|	-a_xx2/denom ~ -(lambda*a_pp1+a_px1)*a_px2/denom^2 ~ 0 ~ -gammaf*a_px2/denom ~ -gammaf*a_xx2/denom ~ 0 ~ (1-gammaf*a_pp1)/denom;
	avJac[0] = vec(mJac);
	return !isnan(mJac);
}
objective(const vp, const adFunc, const avScore, const amHessian)
{
	decl isol, mA;		// restricted A
	isol = getsolution(vp, &mA);
	if(isol == 0){
		adFunc[0] = -9999;
		return 0;		// if solution failed, return zero.
	}
	decl arg = mA-g_A;
	adFunc[0] = -trace(g_OmegaInv*arg'g_XX*arg);
//	adFunc[0] = -double(vec(arg)'(g_OmegaInv**g_XX)*vec(arg));
	if(avScore){
		decl mJac;
		JacvecAwrtTheta(&mJac,vp);
		avScore[0] = -2*shape(mJac,sizeof(mJac)/sizeof(vp),sizeof(vp))'(g_OmegaInv**g_XX)*vec(arg);
	}
	return !isnan(adFunc[0]);
}		
	

main()
{
	decl time = timer();
	/* Load data */
	decl data = new Database();
	data.Load("nkpc_20121219.xls");							// load data
    data.Deterministic(FALSE);

	/* Select regressors for VAR */
	decl Y_VAR = 0, X_VAR = 1, share = "lsnfb";				
	decl inflation = "inf_pgdp";				
	data.Select(Y_VAR, {inflation, 0, 0, share, 0, 0});
	data.Select(X_VAR, {inflation, 1, 1, share, 1, 1, inflation, 2, 2, share, 2, 2});
	data.SetSample(1960,1,2011,4);				// select sample

	/* Estimate VAR(2)	*/
	decl mY = data.GetGroup(Y_VAR);				// regressand
	decl mX = data.GetGroup(X_VAR);				// regressors (lags)
	mY -= meanc(mY);	mX -= meanc(mX);		// demean data
	decl T = sizeof(mY);						// sample size
	decl Ahat, xxinv, xx;
	olsc(mY, mX, &Ahat, &xxinv, &xx);
	g_XX = xx; 
	decl Sigma_hat = variance(mY - mX*Ahat);
	g_OmegaInv = invertsym(Sigma_hat);
	decl XXinv_ch = choleski(xxinv);

	decl lambda, gammaf, gammab;				// semistructural params
	decl calvo = 0.6, varrho = 0.5;				// means of Del Negro et al (2007) Beta priors
	decl beta = .99;							// discount factor
	decl rho1 = 0.9, rho2 = -0.8, phi1 = 0, phi2 = 0;	// reduced form parameters for x that lead to strong ID
	[lambda, gammaf, gammab] = structural2semi(calvo|beta|varrho);

	decl fsave = TRUE;							// if TRUE, save prior/posterior draws and estimated densities
	
	decl A0, vp0;
	decl Aminn = unit(2)|zeros(2,2);				// Minnesota prior
 	vp0 = lambda|gammaf|gammab|phi1|rho1|phi2|rho2;

	decl asprior = {"Jeffreys","Minnesota"};

	for(decl iprior = 0; iprior < sizeof(asprior); ++iprior){

	/* Set g prior for A  */
	decl g = iprior ? 1/2 : 0;					// Minnesota prior with g=1/2, else Jeffreys has g=0.
	decl h = int(max(g*T,2));					// for diffuse prior, set the prior Wishart df's to 2 (minimum to draw nonsingular matrix)

	decl W0, xx0inv, xx0;
	xx0inv = xx0 = unit(4);						// initialize, this will be overridden if g>0
	W0 = unit(2);								// initialize, this will be overridden if g>0
	decl W0inv_ch = choleski(invertsym(W0));	// choleski factor to draw inverse wishart posterior
	decl C0		 = invert(W0inv_ch);			// you need its inverse

	println("\nResults with ", asprior[iprior], " prior.\n");
	if(g)
		println("g = ", g);
		
	A0 = Aminn;
	if(iprior)
		println("Prior mean of A", A0);
	decl At	  = 1/(1+g)*(g*A0+Ahat);					// posterior mean of A

	decl Ohat = T*Sigma_hat;							// sum of squared errors
	decl Wt	  = W0 + Ohat + g/(1+g)*(A0-Ahat)'xx*(A0-Ahat);

	decl Winv_ch = choleski(invertsym(Wt));				// choleski factor to draw inverse wishart posterior
	decl C		 = invert(Winv_ch);						// you need its inverse

	oxwarning(0);					// disable ox warnings
		
	format(1000);

	decl aspar = {"lambda","gammaf","gammab"};

	decl nrep = 1e5;
	decl mstructural = constant(.NaN, nrep,sizeof(aspar));	// store draws of structural parameters
	decl vmu = constant(.NaN,nrep,1);						// store values of misspecification parameter mu
	decl mw	   = constant(.NaN, nrep,3);					// store values of w1,w2
	decl vWeight  = constant(.NaN, nrep,1);					// store values of weights 
	decl vresult  = constant(.NaN, nrep,1);					// store convergence result 

	decl mstructural0 = constant(.NaN, nrep,sizeof(aspar));	// store prior draws of structural parameters
	decl vmu0 = constant(.NaN,nrep,1);						// store prior values of misspecification parameter mu
	decl mw0   = constant(.NaN, nrep,3);					// store values of w1,w2
	decl vWeight0 = constant(.NaN, nrep,1);					// store prior values of weights
	decl vresult0  = constant(.NaN, nrep,1);				// store prior convergence result 

	decl iseed = -1;
	ranseed(iseed);											// set initial seed

	decl sfile = sprint("estimate_", asprior[iprior]);			// file name to store results

	decl in0 = 0, in1 = 0, ntotal = 10*nrep;

	decl fuse_t_dist = TRUE, dft = 3;
	decl dbound = 100, vLo = -dbound*ones(7,1), vHi = -vLo;		// bounds for restricted optimization
	for(decl i = 0; i < ntotal; ++i)
	{
		/* sample reduced-form parameters	*/
		decl argA = fuse_t_dist ? rant(4,2, dft) : rann(4,2);

		for(decl post = iseq(g,0); post < 2; ++post)			// do not draw prior if diffuse (g=0)
		{
			if(post == 0 && in0 >= nrep || post == 1 && in1 >= nrep)
				continue;
			decl arg = ranwishart(post*T+h, 2);					// draw standard inverse Wishart
			decl omega = post ? C*invertsym(arg)*C'
							: C0*invertsym(arg)*C0';			// draw of omega from posterior IW(Wt, T+h)
			decl omega_ch = choleski(omega);					// obtain choleski factor of Omega 
			g_OmegaInv = Winv_ch'arg*Winv_ch;					// store Omega^(-1)

			g_A = post ? At + XXinv_ch*argA*omega_ch'/(1+g)		// posterior draw of A
						: A0 + XXinv_ch*argA*omega_ch'/g;		// prior draw 
					
			decl OmegXXinv = (omega_ch*omega_ch')**xxinv;
		
			decl OmegInvXX = g_OmegaInv**g_XX;

			/* obtain structural parameters by minimizing objective	*/	
			decl dobj, vp = vp0;							// initialize at truth
			decl ires = MaxBFGS(objective, &vp, &dobj, 0, 0);
//			if(any(fabs(vp) .> dbound))
//				ires = MaxSQPF(objective, &vp, &dobj, 0, 0, <>, <>, vLo, vHi);

			if(ires > MAX_WEAK_CONV)						// skip draws for which you failed to get the minimum
				continue;

			/*	Get Jacobian of A(theta) and its orthogonal complement */
			decl mJac, H, mU, mV, mW;				 	// Jacobian of the transformation A(theta), and its orthogonal complement
			JacvecAwrtTheta(&mJac,vp);
			mJac = shape(mJac, 8,7);					// Jacobian is returned vectorized, reshape into 8x7 matrix
			decsvd(mJac*mJac',&mU,&mW,&mV);						// perform sing. value decomp of J*J'
			H = mU[][columns(mJac):];							// H is matrix of eigenvectors corresponding to 0 singular values.
			decl mscale = H'OmegXXinv*H;						// if H is vector, this is a scalar
			H = H*choleski(invertsym(mscale));					// standardize
			decl F = mJac'OmegInvXX*mJac;						// quantity used in Jeffreys prior
	
			decl G, iresJ;
			iresJ = NumJacobian(JacvecAwrtTheta, vp, &G);		// Second derivative of A(theta)
			decl Ares;		getsolution(vp, &Ares);				// obtain A(theta)
			decl mu = H'vec(g_A-Ares);
	
			/*	Get ratio of Jacobians	*/
			decl arg1 = fabs(determinant(F - (unit(sizeof(vp))**(mu'*H'))*G));
			if(isfeq(arg1,0)) 									// if this is too close to zero, ratio will not be accurately computed
				continue;	
			decl w1 = fabs(determinant(F))/arg1;

			decl muhat 	= post ? H'vec(Ahat-Ares) : H'vec(A0-Ares);		// misspecification term
			decl z = post ? g_A - Ahat : g_A - A0;						// argument in likelihood
			decl w2 = trace((mu-muhat)'(mu-muhat)) + trace(g_OmegaInv*z'g_XX*z);
			if(!fuse_t_dist)											
				w2 -= trace(g_OmegaInv*z'g_XX*z);						// divide by Gaussian approximating density
			w2 = exp(-1/2*(post+g)*w2);									// divide by variance and exponentiate
			decl w3 = fuse_t_dist ? (post+g)*(trace(g_OmegaInv*z'g_XX*z)/dft+1)^((dft+sizerc(g_A))/2) : 1;	   // if approximating dens is t, divide by that instead
//			if(w2*w3 > nrep)
//				println("Iteration ", i, " produced large w2 = ", w2, " at parameters", vp', g_A);
			
			decl weight = w1*w2*w3;										// overall weight factor
				
			if(isfeq(weight,0) || isnan(weight))
				continue;
		
			if(post){												   // store posterior draws
				vWeight[in1][] = weight;								
				vmu[in1][] = mu;											
				mstructural[in1][] = vp[:2]';							
				vresult[in1][] = ires <= MAX_WEAK_CONV;
				mw[in1][] = w1~w2~w3;
				++in1;
				if(fsave && in1 == nrep){
					decl dbase = new Database();
					dbase.Create(1, 1, 1,sizeof(mstructural), 1);
					dbase.Append(vmu~mstructural, {"mu"}~aspar);
					dbase.Save(sprint(sfile,"_post.csv"));
					delete dbase;
				}
			}
			else{													   // store prior draws
				vWeight0[in0][] = weight;									
				vmu0[in0][] = mu;									
				mstructural0[in0][] = vp[:2]';						
				vresult0[in0][] = ires <= MAX_WEAK_CONV;
				mw0[in0][] = w1~w2~w3;
				++in0;
				if(fsave && in0 == nrep){
					decl dbase = new Database();
					dbase.Create(1, 1, 1,sizeof(mstructural0), 1);
					dbase.Append(vmu0~mstructural0, {"mu"}~aspar);
					dbase.Save(sprint(sfile,"_prior.csv"));
					delete dbase;
				}
			}
		}
		if( ( ((g==0) || (in0 == nrep)) && in1 == nrep) || i == ntotal-1){									// if you've achieved required # of draws from each distn, break out of loop
			println("Total number of draws: ", i+1);
			println("Total prior draws: ", in0);
			println("Total posterior draws: ", in1);
			break;
		}
	}
	if(fsave){
		for(decl post = !g; post < 2; ++post){
			decl dbase = new Database();
			dbase.Create(1, 1, 1,sizeof(post ? mstructural : mstructural0), 1);
			if(post){
				dbase.Append(vmu~vWeight~mw~mstructural, {"mu","weights", "w1","w2","w3"}~aspar);
				dbase.SaveIn7(sprint(sfile,"_post.in7"));
			}
			else{
				dbase.Append(vmu0~vWeight0~mw0~mstructural0, {"mu","weights", "w1","w2","w3"}~aspar);
				dbase.SaveIn7(sprint(sfile,"_prior.in7"));
			}
			delete dbase;
		}
	}

	/*	Clear missing numbers	*/
	decl all= deleter(vWeight~vmu~mstructural);
	vWeight = all[][0];
	vmu 	= all[][1];
	mstructural = all[][2:];
	if(!iseq(g,0)){
		all 	= deleter(vWeight0~vmu0~mstructural0);
		vWeight0 	 = all[][0];
		vmu0 		 = all[][1];
		mstructural0 = all[][2:];
	}
	/*	cap weights -- to remove .Inf	*/
	decl cap = .Inf;
	vWeight = vWeight .> cap .? cap .: vWeight;
	vWeight0 = vWeight0 .> cap .? cap .: vWeight0;
	
	/*  Compute summary statistics	*/
	decl totweight = sumc(vWeight);
	decl totweight0 = sumc(vWeight0);

	vWeight /= totweight;				// normalize weights to sum to 1.
	vWeight0 /= totweight0;

	decl meanpost = mstructural'vWeight;
	decl meanprior = mstructural0'vWeight0;

	decl sdpost = sqrt(sqr(mstructural)'vWeight-sqr(meanpost));
	decl sdprior = sqrt(sqr(mstructural0)'vWeight0-sqr(meanprior));

	/* quantiles	*/
	decl iqr90post = constant(.NaN, sizeof(aspar),2);
	decl iqr90prior = constant(.NaN, sizeof(aspar),2);
	decl iqr99post = constant(.NaN, sizeof(aspar),2);
	decl iqr99prior = constant(.NaN, sizeof(aspar),2);
	for(decl i = 0; i < sizeof(aspar); ++i){
		for(decl post = !g; post < 2; ++post){
			decl vordered = sortbyc(post ? mstructural[][i]~vWeight : mstructural0[][i]~vWeight0,0);		// sort parameters in ascending order
			decl cumWeight = cumulate(vordered[][1]);
			decl idx5 = max(sumc(cumWeight .<= .05)-1,0);						// index position of 5th percentile
			decl idx95 = max(sumc(cumWeight .<= .95)-1,0);						// index position of 95%
			decl idx05 = max(sumc(cumWeight .<= .005)-1,0);						// index position of .5%
			decl idx995 = max(sumc(cumWeight .<= .995)-1,0);					// index position of 99.5%
			if(post){
				iqr90post[i][] = vordered[idx5][0]~vordered[idx95][0];
				iqr99post[i][] = vordered[idx05][0]~vordered[idx995][0];
			}
			else{
				iqr90prior[i][] = vordered[idx5][0]~vordered[idx95][0];
				iqr99prior[i][] = vordered[idx05][0]~vordered[idx995][0];
			}
		}
	}
	if(!iseq(g,0))					// print prior statistics if not diffuse
		println("Prior", "%r", aspar, "%c", {"mean", "sd", "5\%", "95\%", ".5\%", "99.5\%"}, meanprior~sdprior~iqr90prior~iqr99prior); 

	println("Posterior", "%r", aspar, "%c", {"mean", "sd", "5\%", "95\%", ".5\%", "99.5\%"}, meanpost~sdpost~iqr90post~iqr99post); 
	
	// compute kernel density estimates
	SetDrawWindow(asprior[iprior]);
	decl skern = "g";
	decl hkern = 1.06*nrep^(-1/5);		// bandwidth for Gaussian kernel (times sigmahat)
	decl vparbounds = iprior ? <-0.04,0.04;0.6,2;-0.8,0.4;-5,5> : <-0.07,0.05;1,2.8;-1.5,0;-5,5>;
	decl aspar1 = aspar~{"mu"};
	for(decl post = !g; post < 2; ++post){
		decl iqr99 = post ? iqr99post : iqr99prior;
		iqr99 |= <-.Inf,.Inf>;				// last row for mu, not used
		decl sd = post ? sdpost : sdprior;
		sd |= sqrt(varc(post ? vmu : vmu0));			// last row is for mu, irrelevant
		decl mstruct = post ? mstructural~vmu : mstructural0~vmu0;
		decl vWei = post ? vWeight : vWeight0;
		for(decl i = 0; i < sizeof(aspar1); ++i){
			decl vbounds = vparbounds[i][];
			decl pseudosd = iqr99[i][]*<-1;1>/2;		// 99% IQR is 5.2 st.dev's for the Normal. 
			decl hi = hkern*(min(sd[i],pseudosd));					
			decl vrange = range(vbounds[0], vbounds[1], (vbounds[1]-vbounds[0])/127);
			decl mpdf = new matrix[1][sizerc(vrange)];
			for(decl j = 0; j < sizerc(vrange); ++j){
				mpdf[][j] = i == sizeof(aspar1)-1 ? meanc(denskernel((vrange[j]-mstruct[][i])/hi,skern)/hi) 
						: meanc(vWei.*denskernel((vrange[j]-mstruct[][i])/hi,skern)/hi);
			}
			mpdf = mpdf ./ sumr(mpdf)/hi;
			DrawTitle(i, aspar1[i]);
			DrawXMatrix(i, mpdf, post ? {"posterior"} : {"prior"}, vrange, "", 0, 2+post);
			if(fsave){																		// saves in database file for use in different program (e.g., Matlab in paper)
				decl dbase = new Database();
				dbase.Create(1,1,1,sizerc(vrange),1);
				dbase.Append(vrange'~mpdf',{aspar1[i]}~{post ? "Posterior" : "Prior"});
				dbase.Save(sprint("estimate_", iseq(g,0) ? "Jeffreys_" : "Minnesota_", post ? "post_" : "prior_", aspar1[i],".csv"));
			}
		}
	}
	ShowDrawWindow();
	}
	println("lapsed time: ", timespan(time));
}
